import logging

import pandas as pd

from cumulative.animation import Animation
from cumulative.explore import Explore
from cumulative.geometry import Geometry
from cumulative.options import options
from cumulative.plot import Plot
from cumulative.transforms.apply import Apply
from cumulative.transforms.bin import Bin
from cumulative.transforms.cluster import Cluster
from cumulative.transforms.copy import Copy
from cumulative.transforms.derivate import Derivate
from cumulative.transforms.drop import Drop
from cumulative.transforms.features import Features
from cumulative.transforms.fit import Fit
from cumulative.transforms.integrate import Integrate
from cumulative.transforms.interp import Interp
from cumulative.transforms.scale import Scale
from cumulative.transforms.score import Score
from cumulative.transforms.sequence import Sequence
from cumulative.transforms.sort import Sort
from cumulative.transforms.template import Template
from cumulative.transforms.transition import Transition

log = logging.getLogger(__name__)


class Cumulative:
    def __init__(self, df_raw=None, df=None):

        if df is None:
            df = pd.DataFrame()

        self.df_raw = df_raw
        self.df = df
        self.lineage = []

        #
        self.sequence = Sequence(self)
        self.scale = Scale(self)
        self.derivate = Derivate(self)
        self.integrate = Integrate(self)
        self.fit = Fit(self)
        self.score = Score(self)
        self.copy = Copy(self)
        self.sort = Sort(self)
        self.template = Template(self)
        self.interp = Interp(self)
        self.transition = Transition(self)
        self.cluster = Cluster(self)
        self.bin = Bin(self)
        self.drop = Drop(self)
        self.features = Features(self)
        self.apply = Apply(self)

        self.plot = Plot(self)
        self.anim = Animation(self)
        self.explore = Explore(self)
        self.geometry = Geometry(self)

    def explain(self) -> None:
        for idx, transform in list(enumerate(self.lineage))[::-1]:
            name = transform["name"]
            kwargs = ", ".join([f"{k}={v}" for k, v in transform["kwargs"].items()])
            if transform["dst"] is not None:
                dst = f'-> {transform["dst"]}'
            else:
                dst = ""
            log.info(f"[{idx}] {name}({kwargs}) {dst}")

    def register_transform(self, name, cls):
        self.__dict__[name] = cls(self)

    def track(self, name, dst, kwargs):
        self.lineage.append({"name": name, "dst": dst, "kwargs": kwargs})

    def columns_with_prefix(self, prefix, errors="raise"):
        cols = [col for col in self.df.columns if col.startswith(f"{prefix}.") or col == prefix]
        if len(cols) == 0 and errors == "raise":
            raise Exception(f"No matching columns for prefix '{prefix}'")
        return cols

    def frame(self, src=None, idx=None):
        src = options().default_if_null(src, "transforms.source")
        cols = self.columns_with_prefix(src)
        df = self.df[["idx"] + cols]
        if idx:
            df = df[df.idx == idx]
        return df

    def describe(self, src=None):
        """
        Print basic statistics about the collection in the `src` dimension.
        """

        def min_max_diff(s):
            return f"min={s.min()} max={s.max()} diff={s.max() - s.min()}"

        print(f"Count.......: {len(self.df)}")
        print(f"Length......: {min_max_diff(self.df[f'{src}.len'])}")
        print(f"Index min...: {min_max_diff(self.df[f'{src}.x'].apply(min))}")
        print(f"Index max...: {min_max_diff(self.df[f'{src}.x'].apply(max))}")
        print(f"Value min...: {min_max_diff(self.df[f'{src}.y'].apply(min))}")
        print(f"Value max...: {min_max_diff(self.df[f'{src}.y'].apply(max))}")



import matplotlib
import matplotlib as mpl
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from cumulative.options import options


def supersmooth(x, k=0.5):
    # useful to zoom-in (acceleration) close to 0 and 1 (tending to constant velocity at x=.5)
    # [0,1] -> [0,1]
    # k equal to .0: linear output, no transformation (no acceleration)
    # k equal to .5: sinusoidal curve
    # k close to 1: mostly at .5 (max acceleration close to 0 and 1)

    k = max(k, 0)
    k = min(k, 1 - 1e-5)

    if k == 0:
        return x

    a = 2
    b = -1
    c = 0.5
    d = 0.5
    y = (1 - k) * (a * x + b) / (k - 2 * k * np.abs(a * x + b) + 1) * c + d
    return y


class Figure:
    def __init__(self, x_label="X", y_label="Y", ioff=False):
        mpl.rcParams.update(mpl.rcParamsDefault)
        plt.rcParams["font.family"] = "monospace"
        cmap = matplotlib.colormaps["cool"]

        if ioff:
            plt.ioff()
        else:
            plt.ion()

        fig, ax = plt.subplots(figsize=(5, 5))

        ax.clear()
        ax.set_facecolor("black")
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)

        lim_pad = 0.05
        ax.set_xlim(0 - lim_pad, 1 + lim_pad)
        ax.set_ylim(0 - lim_pad, 1 + lim_pad)

        self.fig = fig
        self.ax = ax
        self.cmap = cmap


class Plot:
    def __init__(self, c):
        self.c = c

    def xrays(self, src=None, figure=None, show=True, alpha=1, ms=1, lw=1, k=60, style="."):
        src = options().default_if_null(src, "transforms.source")
        tmp = options().get("transforms.tmp")
        with options().option_context({"transforms": {"source": tmp, "destination": tmp}}):
            self.c.fit(src=src, method="pchip3", k=k, n_samples=k)
            self.c.plot.scatter(figure=figure, show=show, alpha=alpha, ms=ms, lw=lw, style=style)
            self.c.drop()

    def heatmap(
        self, src=None, figure=None, show=True, ms=4, lw=1, k=60, score="idx", alpha=1, alpha_score=False, style="."
    ):
        src = options().default_if_null(src, "transforms.source")
        tmp = options().get("transforms.tmp")
        tmp_score = f"{tmp}.score"
        with options().option_context({"transforms": {"source": tmp, "destination": tmp}}):
            self.c.fit(src=src, method="pchip3", k=k, n_samples=k)
            self.c.score(src=score, dst=tmp_score).sort(by=tmp_score)
            self.c.plot.scatter(
                figure=figure,
                show=show,
                ms=ms,
                lw=lw,
                alpha=alpha * (1 if not alpha_score else tmp_score),
                score=tmp_score,
                style=style,
            )
            self.c.drop()

    def highways(
        self, src=None, figure=None, show=True, score="idx", style="-", lw=1, ms=1, alpha=1, alpha_score=False
    ):
        src = options().default_if_null(src, "transforms.source")
        tmp = options().get("transforms.tmp")
        tmp_score = f"{tmp}.score"
        with options().option_context({"transforms": {"source": tmp, "destination": tmp}}):
            # TODO: move sort to internals of scatter(), without sorting c.df
            self.c.score(src=score, dst=tmp_score).sort(by=tmp_score)
            self.c.plot.scatter(
                src=src,
                figure=figure,
                show=show,
                score=tmp_score,
                lw=lw,
                ms=ms,
                style=style,
                alpha=alpha if not alpha_score else tmp_score,
            )
            self.c.drop()

    def highlights(self, src=None, figure=None, show=True, score="idx", style="-", lw=1, ms=1, alpha_score=False, k=8):
        self.highways(
            src=src, figure=figure, show=show, score=score, style=style, lw=lw, ms=ms, alpha_score=alpha_score
        )

    def highlight(self, idx=None, src=None, figure=None, show=True):

        if idx is None:
            self.c.plot.scatter(figure=figure, show=show, src=src, style="-", color="white")
            return

        if figure is not None:
            fig = figure
        else:
            fig = Figure()

        self.c.plot.scatter(figure=fig, show=False, src=src, style="-", color="white")

        if not isinstance(idx, list):
            idx = [idx]

        for pos, idx_elem in enumerate(idx):
            c_line = self.c.__class__(df=self.c.df[self.c.df["idx"] == idx_elem])
            c_line.plot.scatter(
                figure=fig,
                show=show if pos == len(idx) - 1 else False,
                src=src,
                alpha=1,
                style="--",
                lw=2,
                color="fuchsia",
            )

    def pixelate(self, src=None, figure=None, show=True, k=30, alpha=0.05, ms=5, style="s"):
        src = options().default_if_null(src, "transforms.source")
        tmp = options().get("transforms.tmp")
        with options().option_context({"transforms": {"source": tmp, "destination": tmp}}):
            self.c.interp(n_samples=k, src=src)
            self.c.bin(n_bins=k)
            self.c.fit(method="pchip3", k=k, n_samples=k)
            self.c.plot.scatter(figure=figure, show=show, alpha=alpha, ms=ms, style=style)
            self.c.drop()

    def scatter(
        self,
        src=None,
        figure=None,
        show=True,
        style=".",
        ms=2,
        lw=1,
        score=None,
        alpha=0.5,
        x_label="x",
        y_label="y",
        color="white",
    ):

        src = options().default_if_null(src, "transforms.source")

        if figure is None:
            figure = Figure(x_label=x_label, y_label=y_label)

        for _, row in self.c.df.iterrows():
            row_color = figure.cmap(row[score]) if isinstance(score, str) else color
            row_alpha = row[alpha] if isinstance(alpha, str) else alpha
            pd.Series(row[f"{src}.y"], index=row[f"{src}.x"]).plot(
                style=style,
                lw=lw,
                ms=ms,
                color=row_color,
                alpha=row_alpha,
                ax=figure.ax,
            )

        if show and plt.isinteractive():
            plt.show()



import logging

import pandas as pd
from tqdm.auto import tqdm

from cumulative.options import options
from cumulative.utils import validate_frame

log = logging.getLogger(__name__)


def process_row(func, row, **kwargs):
    return func(row, **kwargs)


class Transform:
    def __init__(self, c, name=None):
        self.c = c
        self.name = self.__class__.__name__.lower() if not name else name

    def transform_row(self, row):
        return pd.Series()

    def __call__(self, **kwargs):
        tqdm_params = options().get("tqdm")
        tqdm_params["desc"] = self.name
        tqdm.pandas(**tqdm_params)
        # The destination prefix is not required/expected by row transforms,
        # let's drop it if present.

        kwargs["src"] = options().default_if_null(kwargs.pop("src", None), "transforms.source")
        dst = options().default_if_null(kwargs.pop("dst", None), "transforms.destination")

        df = self.apply(**kwargs)

        self.c.track(self.name, dst, kwargs)

        if df is None:
            # Sort
            return self.c

        if isinstance(df, pd.DataFrame):
            df.columns = [f"{dst}.{col}" if col != "idx" else col for col in df.columns]
        elif isinstance(df, pd.Series):
            df = df.rename(dst).to_frame()
        else:
            raise Exception("Invalid argument type")

        validate_frame(df, f"Transform {self.name}")

        drop_cols = self.c.columns_with_prefix(dst, errors="ignore")
        self.c.df = self.c.df.drop(columns=drop_cols, errors="ignore")
        self.c.df = pd.concat(
            [self.c.df, df],
            axis=1,
        )
        return self.c

    def apply(self, **kwargs):
        return self.c.df.progress_apply(lambda row: self.transform_row(row, **kwargs), axis=1)



import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform


class Derivate(Transform):
    def transform_row(self, row, src):

        s = pd.Series(np.gradient(row[f"{src}.y"]), index=row[f"{src}.x"], copy=True)

        attrs = {
            "x_min": s.index.min(),
            "x_max": s.index.max(),
            "y_min": s.min(),
            "y_max": s.max(),
        }

        s -= s.min()
        s /= s.max()
        s = s.values

        attrs["x"] = row[f"{src}.x"]
        attrs["y"] = s

        return pd.Series(attrs)



from mltraq.utils.base_options import BaseOptions


class Options(BaseOptions):
    default_values = {
        "reproducibility": {"random_seed": 123},
        "tqdm": {"disable": False, "leave": False, "delay": 0},
        "transforms": {
            "destination": "base",
            "source": "base",
            "tmp": "temp",
            "sequence": {"attributes": "attrs"},
            "fit": {"curve_fit": {"maxfev": 10000}},
        },
        "warnings": {"disable": True},
        "doc": {"url": "https://elehcimd.github.io/cumulative/"},
    }


def options() -> BaseOptions:
    """
    Returns singleton object of options.
    """

    # In some complex cases (parallel execution of runs, options
    # imported in different ways by different modules), the
    # object is quietly copied, causing errors hard to debug.
    #
    # By always calling options(), this issue disappears and
    # things work as expected.
    return Options.instance()



import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform


class Interp(Transform):
    def transform_row(self, row, src, n_samples=1000, complete=False):

        attrs = {}

        if complete:
            attrs["x"] = np.unique(np.concatenate([row[f"{src}.x"], np.linspace(0, 1, num=n_samples)]))
            attrs["y"] = np.interp(attrs["x"], row[f"{src}.x"], row[f"{src}.y"])
        else:
            attrs["x"] = np.linspace(0, 1, num=n_samples)
            attrs["y"] = np.interp(attrs["x"], row[f"{src}.x"], row[f"{src}.y"])

        return pd.Series(attrs)



from cumulative.transforms.transform import Transform


class Score(Transform):
    def apply(self, src, reverse=False, argsort=False):

        s = self.c.df[src].copy()

        s -= s.min()
        s /= s.max()

        # TODO: in case of nans (happening if s is a constant), it doesn't get detected , no warnings .

        if reverse:
            s = 1 - s

        if argsort:
            s = s.argsort() / len(s)
            s -= s.min()
            s /= s.max()

        return s



import numpy as np
import pandas as pd
import scipy
from tsfel.feature_extraction.features import (
    abs_energy,
    auc,
    autocorr,
    average_power,
    calc_centroid,
    calc_max,
    calc_mean,
    calc_median,
    calc_min,
    calc_std,
    calc_var,
    distance,
    entropy,
    fundamental_frequency,
    human_range_energy,
    interq_range,
    kurtosis,
    max_frequency,
    max_power_spectrum,
    mean_abs_deviation,
    mean_abs_diff,
    mean_diff,
    median_abs_deviation,
    median_abs_diff,
    median_diff,
    median_frequency,
    negative_turning,
    neighbourhood_peaks,
    pk_pk_distance,
    positive_turning,
    rms,
    skewness,
    slope,
    spectral_centroid,
    spectral_decrease,
    spectral_distance,
    spectral_entropy,
    spectral_kurtosis,
    spectral_positive_turning,
    spectral_skewness,
    spectral_slope,
    spectral_spread,
    spectral_variation,
    sum_abs_diff,
    zero_cross,
)

from cumulative.transforms.transform import Transform


def feature_peaks(signal):
    attrs = {}
    peaks = scipy.signal.find_peaks(signal)
    attrs["peaks.count"] = peaks[0].shape[0]

    if peaks[0].shape[0] > 0:
        peaks_prominence = scipy.signal.peak_prominences(signal, peaks[0])
        peaks_width = scipy.signal.peak_widths(signal, peaks[0])
        attrs["peaks.value.avg"] = signal[peaks[0]].mean()
        attrs["peaks.prominence.avg"] = np.mean(peaks_prominence[0])
        attrs["peaks.prominence.sum"] = np.sum(peaks_prominence[0])
        attrs["peaks.width.avg"] = np.mean(peaks_width[0])
    else:
        attrs["peaks.value.avg"] = 0
        attrs["peaks.prominence.avg"] = 0
        attrs["peaks.prominence.sum"] = 0
        attrs["peaks.width.avg"] = 0

    return attrs


def feature_stats(signal):
    return {
        "len": signal.shape[0],
        "max": signal.max(),
        "min": signal.min(),
        "std": signal.std(),
        "mean": signal.mean(),
        "median": np.median(signal),
    }


def feature_tsfel(signal):

    # Features from tsfel. Some features have been ignored due to:
    # - more than one value returned (more work required for integration)
    # - not defined in some cases
    #
    # power_bandwidth
    # spectral_roll_off
    # mfcc
    # wavelet_abs_mean
    # wavelet_energy
    # wavelet_entropy
    # wavelet_std
    # wavelet_var
    # ecdf
    # ecdf_percentile
    # ecdf_percentile_count
    # ecdf_slope
    # fft_mean_coeff
    # hist
    # lpcc

    fs = 1
    return {
        "tsfel_abs_energy": abs_energy(signal),
        "tsfel_auc": auc(signal, fs),
        "tsfel_autocorr": autocorr(signal),
        "tsfel_average_power": average_power(signal, fs),
        "tsfel_calc_centroid": calc_centroid(signal, fs),
        "tsfel_calc_max": calc_max(signal),
        "tsfel_calc_mean": calc_mean(signal),
        "tsfel_calc_median": calc_median(signal),
        "tsfel_calc_min": calc_min(signal),
        "tsfel_calc_std": calc_std(signal),
        "tsfel_calc_var": calc_var(signal),
        "tsfel_distance": distance(signal),
        "tsfel_entropy": entropy(signal),
        "tsfel_fundamental_frequency": fundamental_frequency(signal, fs),
        "tsfel_human_range_energy": human_range_energy(signal, fs),
        "tsfel_interq_range": interq_range(signal),
        "tsfel_kurtosis": kurtosis(signal),
        "tsfel_max_frequency": max_frequency(signal, fs),
        "tsfel_max_power_spectrum": max_power_spectrum(signal, fs),
        "tsfel_mean_abs_deviation": mean_abs_deviation(signal),
        "tsfel_mean_abs_diff": mean_abs_diff(signal),
        "tsfel_mean_diff": mean_diff(signal),
        "tsfel_median_abs_deviation": median_abs_deviation(signal),
        "tsfel_median_abs_diff": median_abs_diff(signal),
        "tsfel_median_diff": median_diff(signal),
        "tsfel_median_frequency": median_frequency(signal, fs),
        "tsfel_negative_turning": negative_turning(signal),
        "tsfel_neighbourhood_peaks": neighbourhood_peaks(signal),
        "tsfel_pk_pk_distance": pk_pk_distance(signal),
        "tsfel_positive_turning": positive_turning(signal),
        "tsfel_rms": rms(signal),
        "tsfel_skewness": skewness(signal),
        "tsfel_slope": slope(signal),
        "tsfel_spectral_centroid": spectral_centroid(signal, fs),
        "tsfel_spectral_decrease": spectral_decrease(signal, fs),
        "tsfel_spectral_distance": spectral_distance(signal, fs),
        "tsfel_spectral_entropy": spectral_entropy(signal, fs),
        "tsfel_spectral_kurtosis": spectral_kurtosis(signal, fs),
        "tsfel_spectral_positive_turning": spectral_positive_turning(signal, fs),
        "tsfel_spectral_skewness": spectral_skewness(signal, fs),
        "tsfel_spectral_slope": spectral_slope(signal, fs),
        "tsfel_spectral_spread": spectral_spread(signal, fs),
        "tsfel_spectral_variation": spectral_variation(signal, fs),
        "tsfel_sum_abs_diff": sum_abs_diff(signal),
        "tsfel_zero_cross": zero_cross(signal),
    }


class Features(Transform):
    def transform_row(self, row, src):

        signal = row[f"{src}.y"]
        attrs = {}
        attrs.update(feature_stats(signal))
        attrs.update(feature_peaks(signal))
        attrs.update(feature_tsfel(signal))

        return pd.Series(attrs)



import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform


class Integrate(Transform):
    def transform_row(self, row, src, scale=False, fillna=False):

        s = pd.Series(np.cumsum(row[f"{src}.y"]), index=row[f"{src}.x"], copy=True)

        attrs = {}

        if scale:
            s -= s.min()
            s /= s.max()
            if fillna:
                s = s.fillna(0)

        attrs["x"] = row[f"{src}.x"]
        attrs["y"] = s.values

        return pd.Series(attrs)



import logging
import warnings

import pandas as pd
from tqdm.auto import tqdm

from cumulative.options import options
from cumulative.transforms.transform import Transform
from cumulative.utils import warn

log = logging.getLogger(__name__)


class SequenceWarning(UserWarning):
    pass


class Sequence(Transform):
    def apply(  # noqa
        self,
        src=None,
        group="group",
        x="x",
        y="y",
        cumsum=False,
        min_len=2,
        min_nunique=2,
        agg=None,
        filter_nochange=True,
        limit=None,
        select=None,
    ):

        self.group_col = group
        self.x = x
        self.y = y

        def build_sequence(df):
            s = pd.Series(df[self.y].values, index=df[self.x].values).rename(df.name).astype(float)

            # ensure that, in case of same index, we aggregate the values.
            # pd.Series([0, 2, 2, 4, 5], index=[0,1,2,2,4]) -> pd.Series([0, 2, 6, 5], index=[0,1,2,4])
            s = s.groupby(s.index).sum()

            if cumsum:
                s = s.cumsum()

            attrs_dict = df.groupby(lambda x: True).agg(agg).iloc[0].to_dict() if agg else {}

            if filter_nochange:
                # Drop values that didn't change from prior value, retaining only points with variations.
                s = s[s != s.shift(1)]

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)

                return pd.Series(
                    {
                        "x": s.index.values,
                        "y": s.values,
                        "len": s.values.shape[0],
                        "min": s.min(),
                        "max": s.max(),
                        "sum": s.sum(),
                        "len_positive": s[s > 0].values.shape[0],
                        "len_negative": s[s < 0].values.shape[0],
                        "len_zero": s[s == 0].values.shape[0],
                        "attributes": attrs_dict,
                    }
                )

        tqdm_params = options().get("tqdm")
        tqdm_params["desc"] = "sequence"
        tqdm.pandas(**tqdm_params)

        if not isinstance(self.c.df_raw, pd.DataFrame):
            raise RuntimeError("df_raw is not a valid Pandas DataFrame")

        df = self.c.df_raw.sort_values(by=[self.x, self.y]).groupby(self.group_col).progress_apply(build_sequence)
        df = df.reset_index(names="name").reset_index(names="idx")

        if min_len:
            m = df["x"].apply(len) < min_len
            if m.sum() > 0:
                warn(f"Ignoring {m.sum()} rows (filter: min_len={min_len})", category=SequenceWarning, stacklevel=1)
                df = df[~m]

        if min_nunique:
            m = df["y"].apply(lambda a: pd.Series(a).nunique()) < min_nunique
            if m.sum() > 0:
                warn(
                    f"Ignoring {m.sum()} rows (filter: min_nunique={min_nunique})",
                    category=SequenceWarning,
                    stacklevel=1,
                )

                df = df[~m]

        if agg:
            prefix = f"{options().get('transforms.sequence.attributes')}."
            df = pd.concat([df, pd.json_normalize(df["attributes"]).add_prefix(prefix)], axis=1)

        df = df.drop(columns=["attributes"])

        if select is not None:
            df = select(df)

        if limit is not None:
            df = df.head(limit)

        df = df.reset_index(drop=True)

        return df



import logging

import pandas as pd

from cumulative.transforms.transform import Transform

log = logging.getLogger(__name__)


class ScaleWarning(UserWarning):
    pass


class ExceptionScaler(Exception):
    pass


class Apply(Transform):
    def transform_row(self, row, src, func=None):

        s = pd.Series(row[f"{src}.y"], index=row[f"{src}.x"], copy=True)
        s = func(s)

        attrs = {
            "x": s.index.values,
            "y": s.values,
        }

        return pd.Series(attrs)



from functools import partial

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from IPython.display import HTML, display
from matplotlib.animation import FuncAnimation
from tqdm.auto import tqdm

from cumulative.options import options
from cumulative.plot import Figure


class Animation:
    def __init__(self, c):
        self.c = c

    def render(self, func, n_frames=10, interval=20, show_percentage=True, html5=False):
        figure = Figure(ioff=True)

        mpl.rcParams["animation.embed_limit"] = 2**128

        pbar = tqdm(total=100, desc="animation", **options().get("tqdm"))

        def init_func():
            # We pass to FuncAnimation this empty init function to not call the expensive draw_frame.
            # (If no init_func is passed, func is called initially to clear the figure.)
            pass

        def draw_frame(figure, pct_frames, i_frame):

            pct_frame = pct_frames[i_frame]
            pct_frame_loop = pct_frame * 2

            if pct_frame_loop > 1:
                pct_frame_loop = 2 - pct_frame_loop

            figure.ax.clear()

            func(self.c, figure=figure, i_frame=i_frame, pct_frame=pct_frame, pct_frame_loop=pct_frame_loop)

            if show_percentage:
                figure.ax.text(
                    0.90,
                    0.05,
                    f"{pct_frame * 100:.0f}%".rjust(4),
                    horizontalalignment="center",
                    verticalalignment="center",
                    transform=figure.ax.transAxes,
                    color="white",
                )
            pbar.update(int(pct_frame * 100))

        self.anim = FuncAnimation(
            fig=figure.fig,
            init_func=init_func,
            func=partial(draw_frame, figure, np.linspace(0, 1, num=n_frames)),
            frames=n_frames,
            interval=interval,
        )

        # func is executed as part of .to_html5_video()

        if html5:
            video = self.anim.to_html5_video()
        else:
            video = self.anim.to_jshtml()

        pbar.close()

        html = HTML(video)
        display(html)
        plt.close()



import logging
import warnings
from functools import partial
from itertools import combinations
from math import factorial

import numpy as np
import pandas as pd
import scipy.stats as st
from scipy.interpolate import PchipInterpolator
from scipy.optimize import curve_fit
from scipy.special import betainc

from cumulative.options import options
from cumulative.transforms.transform import Transform
from cumulative.utils import warn

log = logging.getLogger(__name__)


class FitWarning(UserWarning):
    pass


class ExceptionInvalidMethod(Exception):
    pass


def fit_xy_cdf(idx, x, y, distribution_name="beta", loc=0, scale=1, optimize_loc_scale=True, map_params=None):
    # Seed the random number generator for reproducibility
    np.random.seed(options().get("reproducibility.random_seed"))

    distribution = getattr(st, distribution_name)

    p0 = [1] * len(distribution._shape_info())
    if optimize_loc_scale:
        p0 += [loc, scale]

    def cdf(x, *args):
        if optimize_loc_scale:
            param_loc = args[-2]
            param_scale = args[-1]
            params = args[:-2]
        else:
            param_loc = loc
            param_scale = scale
            params = args
        return distribution.cdf(x, *params, loc=param_loc, scale=param_scale)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        popt, _ = curve_fit(
            cdf,
            tuple(x),
            tuple(y),
            check_finite=True,
            nan_policy="raise",
            p0=p0,
            maxfev=options().get("transforms.fit.curve_fit.maxfev"),
        )

    attrs = {}
    if map_params:
        for k, v in zip(map_params, popt):
            attrs[f"model.params.{k}"] = v
    else:
        attrs["model.params"] = popt
    attrs["model.name"] = "cdf"
    attrs["model.distribution"] = distribution_name
    attrs["model.size"] = len(popt)
    return attrs, lambda x: cdf(x, *popt)


def fit_xy_interp(idx, x, y, k=2):

    # Percentile-based Linear Interpolation (PLI)
    x_model = np.linspace(0, 1, num=k + 2)
    y_model = np.interp(x_model, x, y)

    def f(x):
        return np.interp(x, x_model, y_model)

    attrs = {}
    attrs["model.name"] = "interp"
    attrs["model.params"] = (x_model, y_model)
    attrs["model.size"] = k

    return attrs, f


def fit_xy_pchip0(idx, x, y):
    """
    PCHIP 1-D monotonic cubic interpolation on all data points.
    """

    # possible source of warnings for PCHIP:
    # https://stackoverflow.com/questions/14461346/python-pchip-warnings

    # Seed the random number generator for reproducibility
    np.random.seed(options().get("reproducibility.random_seed"))

    attrs = {}
    attrs["model.name"] = "pchip0"
    attrs["model.size"] = (len(x) - 2) * 2

    return attrs, PchipInterpolator(x, y)


def fit_xy_pchip1(idx, x, y, k=1):
    """
    PCHIP 1-D monotonic cubic interpolation on k data points with equi-distant indexes.

    k=1 utilises a memory budget of 2 values.
    """

    # Seed the random number generator for reproducibility
    np.random.seed(options().get("reproducibility.random_seed"))

    # Take `k+2` equidistant indexes with no duplicates including extremes [0,N].
    # With k=1, we obtain [0, median_index, N].
    indexes = np.unique(np.round(np.linspace(0, len(x) - 1, k + 2)).astype(int))

    # With k=1, x_model and x_model will include the median index with X and Y values.
    # These two values are the model budget (extremes are always [0,1]->[0,1] and don't utilise any memory budget).
    # Therefore, with k=1, pchip1 utilises the same memory budget of the Beta distribution with parameters (a,b).
    x_model = x[indexes]
    y_model = y[indexes]

    attrs = {}
    attrs["model.name"] = "pchip1"

    attrs["model.params"] = (x_model, y_model)
    attrs["model.size"] = (len(x_model) - 2) * 2

    return attrs, PchipInterpolator(x_model, y_model)


def fit_xy_pchip2(idx, x, y, k=1, max_combinations=1000, strict=True):
    """
    PCHIP 1-D monotonic cubic interpolation on k optimal data points, found by brute-forcing
    all combinations an picking the combination with the lowest RMSE.
    Similar to PCHIP1, but with an optimal selection of the data points.

    Improvement from PCHIP1: We ignore easy-to-fit data points,
    allocating the memory budget to the most demanding regions.

    k=1 utilises a memory budget of 2 values.
    """

    # Seed the random number generator for reproducibility
    np.random.seed(options().get("reproducibility.random_seed"))

    # Given a budget of k values, we want to determine which are the best k/2 data points to memorise (x,y pairs).

    # The number of combinations for large k becomes quickly intractable. With k sufficiently low, it is always
    # guaranteed to not contain too many combinations. As a solution, we lower k s.t. we bound the maximum number
    # of combinations we test.

    # Number of combinations of sets of K non-repeating characters out of N characters in order: N!/((N-k)!)
    n = len(x)
    n_combinations = factorial(n) / factorial(n - k)
    if k > 1 and n_combinations > max_combinations:
        # If k == 1, attempt to run on all combinations.
        if strict:
            raise RuntimeError(f"{__name__}: idx={idx}: too many combinations for k={k} ({n_combinations})")
        else:
            warn(
                f"idx={idx}: too many combinations for k={k} ({n_combinations}), reducing to k={k - 1}",
                category=FitWarning,
                stacklevel=1,
            )
            return fit_xy_pchip2(idx, x, y, k=k - 1, max_combinations=max_combinations, strict=strict)

    # List of all subsets of indexes, including first and last, with k intermediate indexes
    z = [[0] + list(a) + [x.shape[0] - 1] for a in list(combinations(range(1, x.shape[0] - 1), k))]

    if len(z) == 0:
        if strict:
            # If there are no viable combinations (eg, sequences of length 2 where x=[0,1] and y=[0,1]),
            # either we raise an exception of fall back to use all the data points we have as best effort approach.
            raise RuntimeError(f"{__name__}: idx={idx}: not enough data points for k={k}")
        else:
            z = [list(range(x.shape[0]))]
            warn(
                f"sequence idx={idx} length is {x.shape[0]} and k={k} (not enough data points), falling back to z={z}",
                category=FitWarning,
                stacklevel=1,
            )

    # Let's find the best performing combination by testing them all.
    best_rmse = None
    for i in z:
        x_ = x[i]
        y_ = y[i]
        rmse = np.sqrt(np.mean((PchipInterpolator(x_, y_)(x) - y) ** 2))
        if not best_rmse or rmse < best_rmse:
            best_rmse = rmse
            best_i = i

    x_model = x[best_i]
    y_model = y[best_i]

    attrs = {}
    attrs["model.name"] = "pchip2"
    attrs["model.params"] = (x_model, y_model)
    attrs["model.params.index"] = best_i
    attrs["model.size"] = (len(x_model) - 2) * 2

    return attrs, PchipInterpolator(x_model, y_model)


def fit_xy_pchip3(idx, x, y, k=2, percentiles=None, store_percentiles=False):
    """
    PCHIP 1-D monotonic cubic interpolation on k equi-distant (or custom) percentiles.
    1. First, we fit a PCHIP model with all data points (x,y).
    2. We then extract the (x_hat,y_hat) pairs at percentiles x_hat (with k: equi-distant).
    3. We fit a new PCHIP model on (x_hat,y_hat), and return it.

    k=2 utilises a memory budget of 2 values.

    Improvement from PCHIP2: It does not suffer from the explosion of
    possible combinations of selected data points, it relies on a
    PCHIP model that passes thru all the data points (x,y),
    it provides equi-distant percentiles controlled by k using all memory budget for Y values,
    it produces easier-to-compare models (all models are aligned on the same X percentiles).
    Further, it allows to specify in which areas of the curve to invest more memory budget
    (eg., specifying percentiles=[.8,.9], we use these two X values to fit the final PCHIP model.)
    Compared to PCHIP2, it does not necessarily pass thru any of the input data points.
    """

    # Seed the random number generator for reproducibility
    np.random.seed(options().get("reproducibility.random_seed"))

    m = PchipInterpolator(x, y)

    if percentiles:
        # use explicit list of percentiles, giving control on which area we want higher accuracy.
        x_model = [0] + percentiles + [1]
    else:
        # +2 to account for extreme values, that don't utilise memory budget [0,1] -> [0,1].
        x_model = np.linspace(0, 1, k + 2)
    y_model = m(x_model)

    attrs = {}
    attrs["model.name"] = "pchip3"
    attrs["model.params.x"] = x_model
    attrs["model.params.y"] = y_model
    attrs["model.size"] = (len(x_model) - 2) * 2

    if store_percentiles:
        for p in range(101):
            attrs[f"model.p{p}"] = m(p / 100)

    # np.random.uniform(low=0, high=1, size=len(x_model)
    return attrs, PchipInterpolator(x_model, y_model)


def fit_xy_pchip4(idx, x, y, k=1):
    """
    PCHIP 1-D monotonic cubic interpolation on k optimized percentiles.
    1. First, we fit a PCHIP model with all data points (x,y).
    2. We find the optimal values for x_hat.
    3. We fit a new PCHIP model on (x_hat,y_hat), and return it.

    k=1 utilises a memory budget of 2 values: x_hat, y_hat.

    Improvement from PCHIP3: x_hat is not fixed anymore, but optimized as well.
    We lose some flexibility, in some cases this translates to higher accuracy.
    This doesn't always results in better performance, likely due to limitations
    of the current optimization process that is constrained and causes
    some issues to curve_
    Compared to PCHIP2, it does not necessarily pass thru any of the input data points.
    """

    # Seed the random number generator for reproducibility
    np.random.seed(options().get("reproducibility.random_seed"))

    m = PchipInterpolator(x, y)

    def valid_xmodel(x_model):
        # PchipInterpolator accepts as input only distinct, increasing values.
        # Further, we append the two percentiles at [0,1].
        return sorted(set([0] + list(x_model) + [1]))

    def f(x, *x_model):
        x_model = valid_xmodel(x_model)
        y_model = m(x_model)
        return PchipInterpolator(x_model, y_model)(x)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        popt, _ = curve_fit(
            f,
            tuple(x),
            tuple(y),
            check_finite=True,
            nan_policy="raise",
            method=None,
            p0=np.random.uniform(0, 1, size=k),
            bounds=[[0] * k, [1] * k],
            maxfev=options().get("transforms.fit.curve_fit.maxfev"),
        )
        x_model = valid_xmodel(popt)
        y_model = m(x_model)

    attrs = {}
    attrs["model.name"] = "pchip4"
    attrs["model.params"] = (x_model, y_model)
    attrs["model.size"] = (len(x_model) - 2) * 2

    return attrs, PchipInterpolator(x_model, y_model)


def fit_xy_beta_pchip(idx, x, y, k=1, **kwargs):
    """
    PCHIP 1-D monotonic cubic interpolation on k equi-distant (or custom) percentiles.
    1. We fit the Beta
    2. We fit using pchip3 on the residuals
    3. We combine the Beta and the pchip3 fit, taking care of returning a monotonic increasing curve.

    Benefits: It relies on the Beta to capture the high-level trend, then relying on PCHIP
    to deformate the Beta in the regions with the highest errors.

    The curve might not be always smooth, with points where the curve remains constant if the sum with
    the residuals would have resulted in a decreasing curve.

    k=2 utilises a memory budget of 4 values (2 for Beta, and 2 for pchip3).
    """

    beta_attrs, beta_fit = fit_xy_betainc(idx, x, y)
    beta_residuals = y - beta_fit(x)
    pchip_attrs, pchip_fit = fit_xy_pchip3(idx, x, beta_residuals, **kwargs)

    def model(x):
        if k == 0:
            return beta_fit(x)
        else:
            return np.maximum.accumulate(beta_fit(x) + pchip_fit(x))

    attrs = {}
    attrs["model.name"] = "beta_pchip"
    attrs["model.params.a"] = beta_attrs["model.a"]
    attrs["model.params.b"] = beta_attrs["model.b"]
    attrs["model.size"] = beta_attrs["model.size"] + pchip_attrs["model.size"]
    return attrs, model


def fit_xy_betainc(idx, x, y, percentage=None, errors="raise", bounds=None):

    # Seed the random number generator for reproducibility
    np.random.seed(options().get("reproducibility.random_seed"))

    if bounds is None:
        # Set bounds to defaults as defined in curve_fit
        bounds = (-np.inf, np.inf)

    try:
        # Handle the special case of all `y` values equal to a constant
        # by raising an exception.
        if len(set(y)) == 1:
            raise RuntimeError("All y values are equal")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            popt, _ = curve_fit(
                lambda x, a, b: betainc(a, b, x),
                tuple(x),
                tuple(y),
                check_finite=True,
                nan_policy="raise",
                method=None,
                bounds=bounds,
                maxfev=options().get("transforms.fit.curve_fit.maxfev"),
            )
            a = popt[0]
            b = popt[1]
    except RuntimeError as e:
        if errors == "diagonal":
            warn(
                f'idx={idx}: Beta fit failed to converge and errors="diagonal", defaulting to a=1 b=1',
                category=FitWarning,
                stacklevel=1,
            )
            a = 1
            b = 1
        else:
            raise e

    if percentage is not None:
        a = 1 + percentage * (a - 1)
        b = 1 + percentage * (b - 1)

    attrs = {}
    attrs["model.name"] = "betainc"
    attrs["model.a"] = a
    attrs["model.b"] = b
    attrs["model.ab_ratio"] = (a + 1) / (b + 1)
    attrs["model.ab_equal"] = 1 / (1 + np.abs(a - b))
    attrs["model.size"] = 2

    return attrs, partial(betainc, a, b)


def fit_xy_curvefit(idx, x, y, func=None, map_params=None, p0=None, bounds=None):

    # Seed the random number generator for reproducibility
    np.random.seed(options().get("reproducibility.random_seed"))

    if func is None:

        def func(x, a, b):
            return betainc(a, b, x)

    if bounds is None:
        # Set bounds to defaults as defined in curve_fit
        bounds = (-np.inf, np.inf)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        popt, _ = curve_fit(
            func,
            tuple(x),
            tuple(y),
            check_finite=True,
            nan_policy="raise",
            method=None,
            bounds=bounds,
            p0=p0,
            maxfev=options().get("transforms.fit.curve_fit.maxfev"),
        )

    attrs = {}
    attrs["model.name"] = "curvefit"
    attrs["model.func"] = func.__name__
    if map_params:
        for k, v in zip(map_params, popt):
            attrs[f"model.params.{k}"] = v
    else:
        attrs["model.params"] = popt
    attrs["model.size"] = len(popt)

    return attrs, lambda x: func(x, *popt)


def fit_xy(idx, x, y, method="betainc", **method_kwargs):  # noqa

    try:
        if method == "betainc":
            return fit_xy_betainc(idx, x, y, **method_kwargs)
        if method == "curvefit":
            return fit_xy_curvefit(idx, x, y, **method_kwargs)
        elif method == "cdf":
            return fit_xy_cdf(idx, x, y, **method_kwargs)
        elif method == "interp":
            return fit_xy_interp(idx, x, y, **method_kwargs)
        elif method == "pchip0":
            return fit_xy_pchip0(idx, x, y, **method_kwargs)
        elif method == "pchip1":
            return fit_xy_pchip1(idx, x, y, **method_kwargs)
        elif method == "pchip2":
            return fit_xy_pchip2(idx, x, y, **method_kwargs)
        elif method == "pchip3":
            return fit_xy_pchip3(idx, x, y, **method_kwargs)
        elif method == "pchip4":
            return fit_xy_pchip4(idx, x, y, **method_kwargs)
        elif method == "beta_pchip":
            return fit_xy_beta_pchip(idx, x, y, **method_kwargs)
        else:
            raise ExceptionInvalidMethod(f"No valid method: '{method}'")
    except (RuntimeError, TypeError, ValueError) as e:
        warn(f"{__name__}: idx={idx}: {e.__class__.__name__}: {e}", category=FitWarning, stacklevel=1)
        attrs = {}
        attrs["model.name"] = method
        return attrs, lambda x: np.full(x.shape, np.nan)


class Fit(Transform):
    def transform_row(self, row, src, method="interp", n_samples=1000, **method_kwargs):

        attrs, func = fit_xy(row["idx"], row[f"{src}.x"], row[f"{src}.y"], method, **method_kwargs)

        attrs["x"] = np.linspace(0, 1, num=n_samples)
        attrs["y"] = func(attrs["x"])
        attrs["y_scale"] = func(row[f"{src}.x"])
        attrs["error.y"] = attrs["y_scale"] - row[f"{src}.y"]
        attrs["error.rmse"] = np.sqrt(np.mean((attrs["error.y"]) ** 2))
        attrs["error.abs.y"] = np.abs(attrs["error.y"])

        attrs["diagonal_dist.abs"] = pd.Series(attrs["y"] - attrs["x"]).abs().mean()

        for prefix in ["error.abs", "error"]:
            # Calculate statistics on error distribution, absolute and signed residuals.
            attrs[f"{prefix}.mean"] = np.mean(attrs[f"{prefix}.y"])
            attrs[f"{prefix}.sum"] = np.sum(attrs[f"{prefix}.y"])
            attrs[f"{prefix}.median"] = np.median(attrs[f"{prefix}.y"])
            attrs[f"{prefix}.std"] = attrs["x"][np.argmax(attrs["y"] > 0.75)]
            attrs[f"{prefix}.min"] = np.min(attrs[f"{prefix}.y"])
            attrs[f"{prefix}.max"] = np.max(attrs[f"{prefix}.y"])
            attrs[f"{prefix}.p25"] = np.percentile(attrs[f"{prefix}.y"], 25)
            attrs[f"{prefix}.p50"] = np.percentile(attrs[f"{prefix}.y"], 50)
            attrs[f"{prefix}.p75"] = np.percentile(attrs[f"{prefix}.y"], 75)

        attrs["p05"] = attrs["x"][np.argmax(attrs["y"] > 0.05)]
        attrs["p25"] = attrs["x"][np.argmax(attrs["y"] > 0.25)]
        attrs["p50"] = attrs["x"][np.argmax(attrs["y"] > 0.50)]
        attrs["p75"] = attrs["x"][np.argmax(attrs["y"] > 0.75)]
        attrs["p95"] = attrs["x"][np.argmax(attrs["y"] > 0.95)]

        return pd.Series(attrs)



import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform


class Bin(Transform):
    def transform_row(self, row, src, n_bins=10):

        y_bins = np.linspace(0, 1, num=n_bins)
        y_hat = y_bins[np.searchsorted(y_bins, row[f"{src}.y"])]
        return pd.Series({"x": row[f"{src}.x"], "y": y_hat}, copy=True)



import pandas as pd

from cumulative.transforms.transform import Transform


class Transition(Transform):
    def transform_row(self, row, src0, src1, percentage=0.5):
        y = row[f"{src0}.y"] * (1 - percentage) + row[f"{src1}.y"] * percentage
        attrs = {"x": row[f"{src0}.x"], "transition.y": y}
        return pd.Series(attrs, copy=True)



import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform


class ExceptionInvalidName(Exception):
    pass


def normalized_tunable_sigmoid(x, k=0.5):
    k = min(k, 1 - 1e-5)
    y = (1 - k) * (2 * x + -1) / (k - 2 * k * np.abs(2 * x + -1) + 1) * 0.5 + 0.5
    return y


class Template(Transform):
    def transform_row(self, row, src, name="diagonal", **kwargs):
        if name == "diagonal":
            y = row[f"{src}.x"]
        elif name == "sigmoid":
            y = normalized_tunable_sigmoid(row[f"{src}.x"], k=kwargs.get(name, 0.5))
        else:
            raise ExceptionInvalidName(f"Invalid value for parameter 'name': '{name}'")

        return pd.Series({"x": row[f"{src}.x"], "y": y})



class Explore:
    def __init__(self, c):
        self.c = c

    def corr(self, col=None, value_abs=True, display=True, sort=False, limit=None):
        numerics = ["int16", "int32", "int64", "float16", "float32", "float64"]
        df = self.c.df.dropna().select_dtypes(include=numerics).corr()
        df = df.dropna(axis=1, how="all")
        df = df.dropna(axis=0, how="all")
        if value_abs:
            df = df.abs()

        if col:
            df = df[[col]]
            df = df[df.index != col]

        if sort:
            df = df.sort_values(col, ascending=False)

        if limit:
            df = df.head(limit)

        if display:
            return df.style.background_gradient(axis=0)
        else:
            return df



import logging
import warnings

import numpy as np

from cumulative.options import options

log = logging.getLogger(__name__)


class ValidateWarning(UserWarning):
    pass


def validate_frame(df, warning_prefix):

    stacklevel = 2

    if len(df) == 0:
        warn(
            f"{warning_prefix}: No rows in data frame",
            category=ValidateWarning,
            stacklevel=stacklevel,
        )
        return

    # Check for invalid values (nans, infs)
    count_cells = df.size
    count_invalid_cells = 0
    for col in df.columns:
        # Look for infs/nans as cell values
        count_invalid_cells += df[col].isin([np.inf, -np.inf, np.nan]).sum()
        # Look for infs/nans inside ndarrays
        count_invalid_cells += df[col].apply(lambda a: isinstance(a, np.ndarray) and (not np.isfinite(a).any())).sum()

    if count_invalid_cells > 0:
        warn(
            f"{warning_prefix}: {count_invalid_cells} ({count_invalid_cells / count_cells * 100:.0f}%) not-finite"
            " values (nans/infs) in data frame",
            category=ValidateWarning,
            stacklevel=stacklevel,
        )

    if len(set(df.columns)) != len(df.columns):
        warn(
            f"{warning_prefix}: Duplicate column names in data frame",
            category=ValidateWarning,
            stacklevel=stacklevel,
        )

    if len(df.columns) == 0:
        warn(
            f"{warning_prefix}: No columns in data frame",
            category=ValidateWarning,
            stacklevel=stacklevel,
        )


def warn(*args, stacklevel=1, **kwargs):
    if not options().get("warnings.disable"):
        warnings.warn(*args, stacklevel=stacklevel, **kwargs)



from cumulative.transforms.transform import Transform


class Drop(Transform):
    def apply(self, src):
        cols = self.c.columns_with_prefix(src, errors="ignore")
        self.c.df = self.c.df.drop(columns=cols)



import numpy as np

from cumulative.options import options


def intersects(x, y, x_min=None, x_max=None, y_min=None, y_max=None):
    # Returns true if any point in the curve x,y matches the constraints
    match = np.full(x.shape[0], True)
    if x_min:
        match = np.logical_and(match, x >= x_min)
    if x_max:
        match = np.logical_and(match, x <= x_max)
    if y_min:
        match = np.logical_and(match, y >= y_min)
    if y_max:
        match = np.logical_and(match, y <= y_max)
    return np.any(match)


class Geometry:
    def __init__(self, c):
        self.c = c

    def intersects(self, src=None, x_min=None, x_max=None, y_min=None, y_max=None):
        src = options().default_if_null(src, "transforms.source")
        matches = self.c.df.apply(
            lambda row: intersects(row[f"{src}.x"], row[f"{src}.y"], x_min, x_max, y_min, y_max), axis=1
        )
        return self.c.df[matches][["idx"]]



import logging

import pandas as pd

from cumulative.transforms.transform import Transform
from cumulative.utils import warn

log = logging.getLogger(__name__)


class ScaleWarning(UserWarning):
    pass


class ExceptionScaler(Exception):
    pass


class Scale(Transform):
    """
    MinMax normalizer on x,y axes to the unit interval [0,1].
    If the sequence contains less than two distinct values, all values are set to `default_value`.
    """

    def transform_row(self, row, src, default_value=0):

        idx = row["idx"]
        s = pd.Series(row[f"{src}.y"], index=row[f"{src}.x"], copy=True)

        attrs = {
            "x_min": s.index.min(),
            "x_max": s.index.max(),
            "y_min": s.min(),
            "y_max": s.max(),
        }

        if s.shape[0] == 0:
            raise ExceptionScaler(f"idx={idx}: Length is zero")

        if s.shape[0] == 1:
            warn(
                f"idx={idx}: Length equal to 1, defaulting `y` to {default_value}", category=ScaleWarning, stacklevel=1
            )
            s[:] = default_value
        elif s.nunique() < 2:
            warn(
                f"idx={idx}: Less than 2 distinct `y` values, defaulting `y` to {default_value}",
                category=ScaleWarning,
                stacklevel=1,
            )
            s[:] = default_value
        else:
            s -= s.min()
            s /= s.max()

        if s.shape[0] == 1:
            warn(f"idx={idx}: Length equal to 1, defaulting `x` to zero", category=ScaleWarning, stacklevel=1)
            s.index = [0]
        else:
            if s.index.nunique() < 2:
                warn(
                    f"idx={idx}: Less than 2 distinct `x` values, defaulting to [0,1,...]/N",
                    category=ScaleWarning,
                    stacklevel=1,
                )
                s = s.reset_index(drop=True)
            s.index -= s.index.min()
            s.index /= s.index.max()

        attrs = {
            **attrs,
            **{
                "x": s.index.values,
                "y": s.values,
            },
        }

        return pd.Series(attrs)



from cumulative.transforms.transform import Transform


class Copy(Transform):
    def apply(self, src):
        cols = self.c.columns_with_prefix(src)
        df = self.c.df[cols].copy()
        df.columns = [col.removeprefix(f"{src}.") for col in cols]
        return df
